﻿namespace FibonacciSequence {
    class Program {
        // C++ code from C++20 standard:
        // 
        // Generator<std::uint64_t> fibonacci_sequence(unsigned n) {
        //    if (n == 0)
        //        co_return;

        //    if (n > 94)
        //        throw std::runtime_error("Too big Fibonacci sequence. Elements would overflow.");

        //    co_yield 0;

        //    if (n == 1)
        //        co_return;

        //    co_yield 1;

        //    if (n == 2)
        //        co_return;

        //    std::uint64_t a = 0;
        //    std::uint64_t b = 1;

        //    for (unsigned i = 2; i < n; i++) {
        //        std::uint64_t s = a + b;
        //        co_yield s;
        //        a = b;
        //        b = s;
        //    }
        //}

        #region V0
        // Query: "Translate the following C++ function to C#:"
        // ChatGPT 3.5:

        static async IAsyncEnumerable<ulong> FibonacciSequenceAsync(uint n) {
            if (n == 0)
                yield break;

            if (n > 94)
                throw new InvalidOperationException("Too big Fibonacci sequence. Elements would overflow.");

            yield return 0;

            if (n == 1)
                yield break;

            yield return 1;

            if (n == 2)
                yield break;

            ulong a = 0;
            ulong b = 1;

            for (uint i = 2; i < n; i++) {
                ulong s = a + b;
                yield return s;
                a = b;
                b = s;
            }
        }
        #endregion

        #region V1
        // Query: "Make FibonacciSequenceAsync sync, please"
        // ChatGPT:

        static IEnumerable<ulong> FibonacciSequenceV1(uint n) {
            if (n == 0)
                yield break;

            if (n > 94)
                throw new InvalidOperationException("Too big Fibonacci sequence. Elements would overflow.");

            yield return 0;

            if (n == 1)
                yield break;

            yield return 1;

            if (n == 2)
                yield break;

            ulong a = 0;
            ulong b = 1;

            for (uint i = 2; i < n; i++) {
                ulong s = a + b;
                yield return s;
                a = b;
                b = s;
            }
        }

        // ChatGPT generated for query:
        // "Write a function using FibonacciSequence that will pass its argument N to it,
        //  and then either return the returned sequence or return null,
        //  if FibonacciSequence throws exception because of N too high."

        static IEnumerable<ulong> GetFibonacciSequenceV1(uint n) {
            IEnumerable<ulong> fibonacciSequence = null;

            try {
                fibonacciSequence = FibonacciSequenceV1(n);
            } catch (InvalidOperationException) {
                // If the FibonacciSequence function throws an exception because n is too high,
                // return null instead of the sequence.
                return null;
            }

            return fibonacciSequence;
        }

        static void ChatGPTTestV1() {
            uint n = 100;

            IEnumerable<ulong> fibonacciSequence = GetFibonacciSequenceV1(n);

            if (fibonacciSequence == null) {
                Console.WriteLine($"n is too high: {n}");
            } else {
                foreach (var fib in fibonacciSequence) {
                    Console.WriteLine(fib);
                }
            }
        }

        #endregion

        #region V2
        // Query: "But it never prints "n is too high", instead the exception is thrown in the foreach cycle. What is wrong?"
        // ChatGPT: "You're correct, the InvalidOperationException exception will be thrown when the foreach loop tries to iterate over the sequence, not when the GetFibonacciSequence function is called. To fix this, you can modify the GetFibonacciSequence function to return an empty sequence instead of null when an exception is thrown. Here's the modified code:"

        static IEnumerable<ulong> GetFibonacciSequenceV2(uint n) {
            try {
                return FibonacciSequenceV1(n);
            } catch (InvalidOperationException) {
                // If the FibonacciSequence function throws an exception because n is too high,
                // return an empty sequence instead of null.
                return Enumerable.Empty<ulong>();
            }
        }

        static void ChatGPTTestV2() {
            uint n = 100;

            IEnumerable<ulong> fibonacciSequence = GetFibonacciSequenceV2(n);

            if (!fibonacciSequence.Any()) {
                Console.WriteLine($"n is too high: {n}");
            } else {
                foreach (var fib in fibonacciSequence) {
                    Console.WriteLine(fib);
                }
            }
        }
        #endregion

        #region V3
        // Query: "Now the exception is thrown inside Any(), so "n is too high" is still not printed. Where is the problem?"
        // ChatGPT: "You are correct. The Any() method will still throw an exception if the sequence is empty, and so the "n is too high" message will not be printed. One solution to this is to use a try-catch block around the foreach loop instead of using the Any() method.Here's an updated version of the example code that uses a try-catch block:"

        static void ChatGPTTestV3() {
            uint n = 100;

            IEnumerable<ulong> fibonacciSequence = GetFibonacciSequenceV2(n);

            try {
                foreach (var fib in fibonacciSequence) {
                    Console.WriteLine(fib);
                }
            } catch (InvalidOperationException) {
                Console.WriteLine($"n is too high: {n}");
            }
        }
        #endregion

        static void Main(string[] args) {
            // ChatGPTTestV1();
            ChatGPTTestV2();
            ChatGPTTestV3();

            GenerateAndPrintFibonacciSequence(0);
            GenerateAndPrintFibonacciSequence(2);
            GenerateAndPrintFibonacciSequence(10);
            GenerateAndPrintFibonacciSequence(100);
        }

        static void GenerateAndPrintFibonacciSequence(uint n) {
            var s = CorrectedFibonacciSequence(n);
            Console.WriteLine($"Got sequence for {n} ...");
            PrintAll(s);
        }

        static void PrintAll<T>(IEnumerable<T> list) {
            Console.WriteLine("{");
            foreach (var item in list) {
                Console.WriteLine($"  > {item}");
            }
            Console.WriteLine("}");
        }

        public static IEnumerable<ulong> CorrectedFibonacciSequence(uint n) {
            if (n > 94)
                throw new InvalidOperationException("Too big Fibonacci sequence. Elements would overflow.");

            return CorrectedFibonacciSequenceUnsafe(n);
        }

        private static IEnumerable<ulong> CorrectedFibonacciSequenceUnsafe(uint n) {
            if (n == 0)
                yield break;

            yield return 0;

            if (n == 1)
                yield break;

            yield return 1;

            if (n == 2)
                yield break;

            ulong a = 0;
            ulong b = 1;

            for (uint i = 2; i < n; i++) {
                ulong s = a + b;
                yield return s;
                a = b;
                b = s;
            }
        }

        // Query: Change n parameter of FibonacciSequence to int, and add any additional checks necessary.
        // ChatGPT: After changing the n parameter of FibonacciSequence to int, we need to check for negative values of n and handle them appropriately. One possible solution is to throw an ArgumentOutOfRangeException with a descriptive message.

        public static IEnumerable<ulong> CorrectedFibonacciSequenceV2(int n) {
            if (n < 0)
                throw new ArgumentOutOfRangeException(nameof(n), "The number of elements in the Fibonacci sequence cannot be negative.");

            if (n > 94)
                throw new InvalidOperationException("Too big Fibonacci sequence. Elements would overflow.");

            return CorrectedFibonacciSequenceUnsafeV2(n);
        }

        private static IEnumerable<ulong> CorrectedFibonacciSequenceUnsafeV2(int n) {
            if (n == 0)
                yield break;

            yield return 0;

            if (n == 1)
                yield break;

            yield return 1;

            if (n == 2)
                yield break;

            ulong a = 0;
            ulong b = 1;

            for (uint i = 2; i < n; i++) {
                ulong s = a + b;
                yield return s;
                a = b;
                b = s;
            }
        }

        // Query: "Write detailed XML docs comments for the public FibonacciSequence capturing all its behavior. Functions have code:"
        // ChatGPT:

        /// <summary>
        /// Generates a sequence of Fibonacci numbers up to the given index (n).
        /// </summary>
        /// <param name="n">The number of Fibonacci numbers to generate.</param>
        /// <returns>An enumerable sequence of Fibonacci numbers up to the given index (n).</returns>
        /// <remarks>
        /// The Fibonacci sequence is a sequence of numbers in which each number is the sum of the two preceding numbers, 
        /// usually starting with 0 and 1. However, this function always starts the Fibonacci sequence with 0 and never with 1.
        /// The first two numbers of the Fibonacci sequence are 0 and 1, and each subsequent number is the sum of the 
        /// previous two. For example, a Fibonacci sequence of 6 elements would be: 0, 1, 1, 2, 3, 5.
        /// The function throws an InvalidOperationException if the requested number of elements is greater than 94, 
        /// because the resulting numbers would overflow.
        /// </remarks>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when n is negative.</exception>
        /// <exception cref="InvalidOperationException">Thrown when n is greater than 94.</exception>
        public static IEnumerable<ulong> FibonacciSequence(int n) {
            if (n < 0)
                throw new ArgumentOutOfRangeException(nameof(n), "The number of elements in the Fibonacci sequence cannot be negative.");

            if (n > 94)
                throw new InvalidOperationException("Too big Fibonacci sequence. Elements would overflow.");

            return FibonacciSequenceUnsafe(n);
        }

        /// <summary>
        /// Generates a sequence of Fibonacci numbers up to the given index (n) without checking for overflow.
        /// </summary>
        /// <param name="n">The number of Fibonacci numbers to generate.</param>
        /// <returns>An enumerable sequence of Fibonacci numbers up to the given index (n) without checking for overflow.</returns>
        /// <remarks>
        /// This function generates a sequence of Fibonacci numbers up to the given index (n) without checking for overflow. 
        /// It should only be called from the public FibonacciSequence function, which checks for overflow before calling this function.
        /// </remarks>
        private static IEnumerable<ulong> FibonacciSequenceUnsafe(int n) {
            if (n == 0)
                yield break;

            yield return 0;

            if (n == 1)
                yield break;

            yield return 1;

            if (n == 2)
                yield break;

            ulong a = 0;
            ulong b = 1;

            for (int i = 2; i < n; i++) {
                ulong s = a + b;
                yield return s;
                a = b;
                b = s;
            }
        }

    }
}