﻿using System.Reflection;

namespace BasicReflection {
 
    class A {
        public void Run() {
            Console.WriteLine("A is running ...");
        }

        public string Combine(string s, long value) { return "A:" + s + value; }
    }

    class B {
        public void Run() {
            Console.WriteLine("B is running ...");
        }

        public long Combine(string s, long value) { return value; }

        public int Combine(int a, int b) { return a + b; }
    }

    internal class Program {
        static void Main(string[] args) {
            A a = new A();
            B b = new B();

            RunIt(a);
            RunIt(b);

            Console.WriteLine();

            PrintVia(a);
            PrintVia(b);

            Console.WriteLine();

            PrintViaV2(a);
            PrintViaV2(b);

            Console.WriteLine();

            PrintViaWrongV3(a);
            PrintViaWrongV3(b);
        }

        static void RunIt(object o) {
            MethodInfo? mi = o.GetType().GetMethod("Run");
            if (mi is null) {
                Console.WriteLine("RunIt: duck typing failed, Run not found ...");
            } else {
                Console.Write("RunIt: ");
                mi.Invoke(o, null);
            }
        }

        static void PrintVia(object o) {
            MethodInfo? mi = o.GetType().GetMethod("Combine", new Type[] { typeof(string), typeof(long) });
            if (mi is null) {
                Console.WriteLine("PrintVia: duck typing failed, Combine(string, long) not found ...");
            } else {
                object? result = mi.Invoke(o, new object[] { "Hello", 123L });
                if (result is string s) {
                    Console.WriteLine($"PrintVia: got string '{s}' ...");
                } else {
                    Console.WriteLine($"PrintVia: got {result?.GetType()}: '{result?.ToString()}' ...");
                }
            }
        }

        static void PrintViaV2(object o) {
            MethodInfo? mi = o.GetType().GetMethod("Combine", new Type[] { typeof(string), typeof(long) });
            if (mi is null) {
                Console.WriteLine("PrintVia: duck typing failed, Combine(string, long) not found ...");
            } else {
                object? result = mi.Invoke(o, new object[] { "Hello", 456 });
                if (result is string s) {
                    Console.WriteLine($"PrintVia: got string '{s}' ...");
                } else {
                    Console.WriteLine($"PrintVia: got {result?.GetType()}: '{result?.ToString()}' ...");
                }
            }
        }

        static void PrintViaWrongV3(object o) {
            MethodInfo? mi = o.GetType().GetMethod("Combine");
            if (mi is null) {
                Console.WriteLine("PrintVia: duck typing failed, Combine not found ...");
            } else {
                object? result = mi.Invoke(o, new object[] { "Hello", 789L });
                if (result is string s) {
                    Console.WriteLine($"PrintVia: got string '{s}' ...");
                } else {
                    Console.WriteLine($"PrintVia: got {result?.GetType()}: '{result?.ToString()}' ...");
                }
            }
        }
    }
}